/*
 * icondesk.cc - code for the icons part of the window
 *
 * This file is part of the source distribution of 
 * Jafm - an icon based X11 desktop program 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

#include <qregion.h>
#include <qtstream.h>
#include <qlist.h>
#include <qtimer.h>
#include <qregexp.h>
#include "iconview.hh"
#include "sfm.hh"

#include "math.h"

#include <iostream.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

#include <X11/Xlib.h>
#include <X11/extensions/shape.h>

#define INSET 20

IconDesk::IconDesk (IconView *parent) 
  : QWidget (parent)
{	   
  owner = parent ;
  
  icon_list = new QList<Icon> ;
  icon_list->setAutoDelete (true) ;

  labelEdit = 0 ;
  
  current_dir = 0 ;

  selectedList = new QList<Icon> ;
  dragWidgetList = new QList<DeskIcon> ;
  dragWidgetList->setAutoDelete (true) ;

  offsetx = 0 ; offsety = 0 ;
  vxsize = width () ; vysize = height () ;
  rootDesk = ((IconView *)parent)->getRootDesk() ;
  
  focusIcon = 0 ; // assume that no icon is under pointer
  dragDropIcon = 0 ;
  scale = 6 ;
  setFont (QFont ("Helvetica", appropSize() )) ;  
  setMouseTracking (true) ;

  // timer to check files
  refresh_timer = new QTimer (this) ;
  connect (refresh_timer, SIGNAL  (timeout()),
	   this, SLOT (slot_refresh())) ;

  // signal to tell everyone if an icon is moved or renamed
  // or to let us know same
  connect (this, SIGNAL( broadcast_file_change (const QFileInfo &, const QFileInfo &)),
	   (Sfm *)qApp, SLOT (slot_file_change (const QFileInfo &, const QFileInfo &))) ;
  connect ((Sfm *)qApp, SIGNAL (broadcast_file_change (const QFileInfo &, const QFileInfo &)),
	   this, SLOT (slot_file_change(const QFileInfo &, const QFileInfo &))) ;

} 

void IconDesk::read (const QDir &dir) 
{
  /* read in all files in current directory.
   * can be time consuming.
   */
  
  refresh_timer->stop () ;
  
  if (! QDir::setCurrent (dir.path())) {
    debug ("Couldn't set dir to %s",dir.path()) ;
    refresh_timer->start (10000) ;
    return ;
  }
  dir_last_mod = 0 ; //should set this correctly 
  setCursor (waitCursor) ;
  if (current_dir == 0) {
    // Not sure if I really know how QDir works...
    current_dir = new QDir () ; 
  } 
  *current_dir = dir.absPath () ;
  current_dir->setNameFilter (dir.nameFilter()) ;
  
  const QStrList *fileList = dir.entryList () ;
  CHECK_PTR (fileList) ;
  QStrListIterator it (*fileList) ;
  
  icon_list->clear () ;
  selectedList->clear () ;
  focusIcon = 0 ;
  dragDropIcon = 0 ;

  for ( ; it.current () ; ++it) {
    
    icon_list->append ( new Icon((const char *)it.current(), 0, scale) ) ;
    icon_list->current()->set_accepts_drag_drop (icon_list->current()->is_drag_droppable()) ;
  }
  layoutIcons () ;
  repaint () ;
  setCursor (arrowCursor) ;
  
  // set updates
  refresh_timer->start (1000) ;

}
 
void IconDesk::read (const char *set)
{
  /* read in icons from a set list
   * need not be associated with real files
   */

  char buf[200] ; // bug
  if (current_dir) 
    delete current_dir ;
  current_dir = 0 ;

  int n = strlen (APPLDIR) 
    + strlen (set) ;
  char *fname = new char[n+2] ;
  strcpy (fname, APPLDIR) ;
  strcat (fname,"/") ;
  strcat (fname, set) ;
  
  QFile f (fname) ;
  if (!f.open (IO_ReadOnly) ) {
    debug ("Couldn't open %s", f.name()) ;
  }

  QTextStream stream (&f) ;

  icon_list->clear () ;
  selectedList->clear () ;
  focusIcon = 0 ;
  dragDropIcon = 0 ;

  while (!stream.eof()) {
    stream >> buf ; // must tighten this code up...
    for (int i = 0 ; i < strlen (buf) ; ++i) { // ignore whitespace only words
      if (!isspace (*(buf+i))) 
	goto a ;
    }
    continue ;
  a:
    icon_list->append ( new Icon ( buf, 0, scale, false)) ;
    CHECK_PTR(icon_list->current()) ;
    icon_list->current()->
      set_accepts_drag_drop (icon_list->current()->is_drag_droppable()) ;    
  }
  layoutIcons () ;
}



void IconDesk::paintEvent (QPaintEvent *pe) 
{
  /* repaint ourself */

  // icons first

  QListIterator<Icon> it (*icon_list) ;
  QPainter p ;
  p.begin (this) ;
  p.translate (offsetx, offsety) ;
  QRect r = pe->rect() ;
  r.moveBy (-offsetx, -offsety) ;

  for ( ; it.current () ; ++it) {
    if (it.current()->rect().intersects (r)
	|| it.current()->get_label_rect().intersects (r))
      it.current()->draw (p) ;
  }
  
  // misc highlighting stuff
  p.resetXForm () ;
  QPen pen (colorGroup().dark(), 2) ;
  p.setPen (pen) ;
  p.drawLine (1,1,width(),1) ;
  p.drawLine (1,1,1,height()) ;
  //  pen.setColor (colorGroup().mid()) ;
  // p.setPen (pen) ;
  // p.drawLine (2,height()-1,width()-2,height()-1) ;
  // p.drawLine (width()-1,1,width()-1, height()-2) ;
  

  p.end () ;
}

void IconDesk::drawIcons (const QRect &r, bool erase)
{
  /* draw all icons that lie inside the rectangle r */

  QListIterator<Icon> it (*icon_list) ;
  QPainter p ;
  p.begin (this) ;
  if (erase) p.eraseRect (r) ;
  //  p.translate (offsetx, offsety) ;
  for ( ; it.current () ; ++it) {
    if (it.current()->rect().intersects (r)) 
      it.current()->draw (p) ;
  }
  p.end () ;
}

void IconDesk::single_repaint (const Icon* ic)
{
  QPainter p ;
  p.begin (this) ;
  p.translate (offsetx, offsety) ;
  ((Icon *)ic)->draw (p) ; // cast away const - shouldn't have to
  p.end () ;
}


void IconDesk::layoutIcons () 
{
  
  if (icon_list->isEmpty()) return ; // no icons
  
  QListIterator<Icon> itI (*icon_list) ;
  
  int xp = INSET ; 
  
  // put roughly in a 4 * 3 grid
  // this fails to account for label size however

  int cols =  (int) (sqrt (icon_list->count() * 1.33)) ;
  
  // the height should be the same for each icon on the screen
  int ydel = itI.current()->height() + INSET + INSET/2 ;

  // label height is line spacing for font
  ydel += fontMetrics().lineSpacing() ;
  int xdel ;
  int yp = INSET ;
  int xnum = 1 ;

  int maxx = 0 ;
  for ( ; itI.current () ; ++itI) {
    
    // calc label width
    int labw = fontMetrics().width (itI.current()->get_label()) ;
    xdel = (labw > itI.current()->width()) ? (labw - itI.current()->width())/2 : 0 ;
    xp += xdel ;
    itI.current ()->move (xp, yp) ;
    xp += itI.current ()->width () + xdel + 20 ;
    if (xnum++ > cols) {
      if (xp > maxx) 
	maxx = xp ;
      xp = INSET , yp += ydel ;
      xnum = 1 ;
    }
  }
  vxsize = width() > maxx ? width() : maxx ;
  vysize = height() > yp ? height() : yp ;
  emit resized(vxsize, vysize) ;
}

void IconDesk::glueLabel (Icon *ic)
{
  (void) ic ;
  /*
  QLineEdit *ed = labelList->at ( icon_list->find (ic)) ;
  ed->move (ic->x() + ic->width()/2 -
				ed->width()/2,
				ic->y() + ic->height()) ;
				*/
}

/* ---------------------------------------------------------------------
 * Mouse Handling Code
 * ---------------------------------------------------------------------
 */


void IconDesk::mousePressEvent (QMouseEvent *me)
{
  
  // if the user clicks in our window, we 
  // make ourselves the current dir
  /*
  if (me->button() == RightButton) {
    debug ("%s", (const char *) current_dir->absPath()) ;
    return ;
  }
  */

  if (current_dir) QDir::setCurrent (current_dir->absPath()) ;
   
  pressPoint = me->pos () ;
  if (labelEdit) { 
    delete labelEdit ;
    labelEdit = 0 ;
  }
  
  // okay, was it pressed inside an icon?
  Icon *ic = point_is_in_icon (me->pos()) ;
  if (ic) {

    // unselect all other icons if shift not held
    if ((me->state() & ShiftButton)) {
      select (ic) ;
      // add all selected 
    } else { // just this one
      unselectAll () ;
      select (ic) ;
    }
    return ;
  }
  // what about an icon label?
  ic = point_is_in_icon_label (me->pos()) ;
  if (ic) {
    
    // edit the name
    unselectAll () ;
    select (ic) ;
    labelEdit = new QLineEdit (this, 0) ;
    labelEdit->setText (ic->get_label()) ;
    connect (labelEdit, SIGNAL (returnPressed()),
	     this, SLOT (rename_by_label ())) ;
    QRect r(ic->get_label_rect ()) ;
    r.setRect (r.x()-2, r.y()-2,r.width()+4,r.height()+4) ;
    labelEdit->setGeometry (r) ;
    labelEdit->setFont (QFont ("Helvetica", ic->label_text_size() )) ;
    labelEdit->move (labelEdit->x()+offsetx, labelEdit->y()+offsety) ;
    labelEdit->show() ;
    return ;
  }
    unselectAll () ;
}

void IconDesk::mouseDoubleClickEvent (QMouseEvent *me)
{
  /* when the mouse is doubled clicked inside an icon,
   * we pass it to the iconview widget to open
   * 
   */

  if (selectedList->isEmpty()) {
    // couldn't be inside an icon, or
    // it would have been selected!
    return ;
  }
  
  if (selectedList->count() == 1) {
    
    if (point_is_in_icon (me->pos(), selectedList->first())) {
      if (selectedList->first()->is_dir()) {
	
	emit newDirWindow ((const char *)selectedList->first()->get_full_name()) ;
	//	owner->changeDir ((const char *)selectedList->first()->get_full_name()) ;
	
      } else {
	selectedList->first()->open_action () ;
      }
    }
    return ;
  }
}

void IconDesk::mouseReleaseEvent (QMouseEvent *me)
{
  // if we are dragging widgets, what to do with them.
  if (!(dragWidgetList->isEmpty())) {
    emit dragDropRelease (mapToGlobal(me->pos()), *dragWidgetList) ;
  }
  // unemboss them all
  QListIterator<Icon> it (*selectedList) ;
  for ( ; it.current () ; ++it) {
    it.current()->set_embossed (false) ;
    single_repaint (it.current()) ;
  }
  dragWidgetList->clear () ;
  if (dragDropIcon) {
    dragDropIcon->set_drag_drop (false) ;
    single_repaint (dragDropIcon) ;
    dragDropIcon = 0 ;
  }
}

void IconDesk::mouseMoveEvent (QMouseEvent *me)
{
  /* track movement of mouse, either for dragging 
   * or focus selection 
   */
  
  if (labelEdit) return ; // no dragging whilst renaming
  
  if (me->state() & LeftButton) { // dragging
    
    if (dragWidgetList->isEmpty()) {
      
      if (selectedList->isEmpty ()) {
	return ; // code for selection rectangle
      }
      // if we're not already dragging, we should be
      // add each selected icon to the list
      QListIterator<Icon> it (*selectedList) ;
      for ( ; it.current() ; ++it) {
	
	it.current()->set_embossed (true) ;
	QPoint offset = pressPoint - me->pos() ;
	
	// make the drag widget 
	/*	QWidget *dragWidget = new QWidget (0, 0, WStyle_Customize |
					   WStyle_NoBorder) ;
		XShapeCombineMask (x11Display(), dragWidget->handle(),
			   ShapeBounding, 0, 0, it.current()->get_mask()->handle(),
			   ShapeSet) ; 
			  DragWidget->setBackgroundPixmap (*(it.current()->getSelected())) ; */
	DeskIcon *dragWidget = new DeskIcon (it.current()) ;
	// move it to it's position, offes
	dragWidget->move (mapToGlobal(it.current()->pos()-offset+QPoint (offsetx, offsety))) ;
	dragWidget->show () ;
	
	dragWidgetList->append(dragWidget) ;
      }
    }  else {
      // just move the currently dragged widgets 
      
      QListIterator<DeskIcon> it (*dragWidgetList) ;
      QPoint offset = lastMousePoint - me->pos () ;
      for ( ; it.current() ; ++it) {
	// offset it's position
	it.current()->move ( it.current()->pos() - offset) ;
      }
    }
    lastMousePoint = me->pos () ;
    emit dragPointAt (mapToGlobal(lastMousePoint), *dragWidgetList) ; // let everyone know about it
    return ;
  }
  
  if (focusIcon) { // if icon is in focus, just check it
    
    if (point_is_in_icon (me->pos(), focusIcon)) 
      return ;
    
    focusIcon->drop_focus () ;
    single_repaint (focusIcon) ;
    focusIcon = 0 ;
  } 
  
// otherwise
  Icon *ic = point_is_in_icon (me->pos()) ;
  if (ic) {
    
    ic->take_focus() ;
    single_repaint (ic) ;
    focusIcon = ic ;
  }
}

void IconDesk::checkDragDrop (QPoint &pos)
{
  /* the point p - in parent coords -
   * is inside: check that if it is over a dragdrop point
   */
  QPoint p = mapFromParent (pos) ;
  Icon *ic = point_is_in_icon (p) ;

  if (ic && ic->will_accept_drag_drop()) {
    
    // that icon cannot be one that is being dragged!
    if (selectedList->containsRef (ic))
      return ;
    
    if (dragDropIcon == ic) 
      return ;
    if (dragDropIcon) {
      dragDropIcon->set_drag_drop (false) ;
      single_repaint (dragDropIcon) ;
    }    
    ic->set_drag_drop (true) ;
    dragDropIcon = ic ;
    single_repaint (ic) ;
    return ;
  }
  if (dragDropIcon) {
    dragDropIcon->set_drag_drop (false) ;
    single_repaint (dragDropIcon) ;
    dragDropIcon = 0 ;
  }
}


/* ---------------------------------------------------------------------
 * Drag Drop Code
 * ---------------------------------------------------------------------
 */


void IconDesk::drag_drop_happened (const QPoint &pos, const QList<DeskIcon> &dic_list)
{
  /* 
   * someone has dropped a bunch of icons inside us.
   */
  
  
  QPoint p = mapFromParent (pos) ;
  if (dragDropIcon) {  // if the icons are held above an icon on our window

    //  debug ("drag drop onto %s", dragDropIcon->get_label()) ;
    
    // if it was all just within this window, de-emboss the originals
    if (&dic_list == dragWidgetList) {
      QListIterator<Icon> it (*selectedList) ;
      for (; it.current() ; ++it) 
	it.current()->set_embossed (false) ;
    } 
    
    dragDropIcon->drag_drop_action (dic_list) ; // do it
    unHighlightDragDrop () ;
    
    emit signal_refresh () ; // unnecessary, but can make things look zippier

  } else { 
    
    if (&dic_list == dragWidgetList) {
      // they were ours to begin with -- just changing positions
      QListIterator<Icon> it (*selectedList) ;
      for (; it.current() ; ++it) {
	QPoint delta = p - pressPoint ;
	QRect ur = it.current()->rect() ;
	ur.moveBy (offsetx,offsety) ;
	update (ur.left(), ur.top(), ur.width(), ur.height()) ;
	ur = it.current()->get_label_rect() ;
	ur.moveBy (offsetx,offsety) ;
	update (ur.left(), ur.top(), ur.width(), ur.height()) ;
	it.current()->offset (delta) ; //update generated automagically!
	it.current()->set_embossed (false) ;
      }    
    } else { // they've been given to us

      if (current_dir == 0) {
	debug ("Sorry - dragging new files into Applications window not yet supported") ;
	return ;
      }
    
      // debug ("move them over to this window") ;
      // debug ("move %s", (const char *) dic_list.getFirst()->get_icon()->get_full_name()) ;
      if (current_dir) {
	// debug ("to %s", (const char *) current_dir->absPath()) ;
	// should check for write permissions here
	
	QListIterator<DeskIcon> it (dic_list) ;
	for ( ; it.current() ; ++it) {

	  // construct the new name of the file
	  size_t n = strlen (current_dir->absPath()) ;
	  char *new_name = new char[n + strlen(it.current()->get_icon()->get_label()) + 2] ;
	  strcpy (new_name, current_dir->absPath()) ;
	  new_name[n++] = '/' ;
	  strcpy (new_name+n, it.current()->get_icon()->get_label()) ;
	  //  debug ("new name is %s", new_name) ;
	  
	  if (rename (it.current()->get_icon()->get_label(), new_name)) {
	    perror ("rename") ; 
	    break ; 
	  } else { // ok, put new icon in the window
	    add_new_icon (it.current()) ;	    
	  }	  
	}
	// send a refresh signal 
	emit signal_refresh () ;
      }
      else
	// debug ("my dir is zero") ;
      return ;
    }
  }
}

void IconDesk::unHighlightDragDrop () 
{
  if (dragDropIcon) {
    dragDropIcon->set_drag_drop (false) ;
    single_repaint (dragDropIcon) ;
    dragDropIcon = 0 ;
  }
}

void IconDesk::rescale (int s)
{
  /* rescale all icons, text, etc
   * to new scale s 
   */
  
  QListIterator<Icon> it (*icon_list) ;
  double del = (double)s/(double) scale ;

  scale = s ;  
  setFont (QFont ("Helvetica", appropSize() )) ;

  for ( ; it.current() ; ++it) {
    
    it.current()->rescale (s) ;
    // how far do we move the icons in or out?
    int qx = (int) ((double)(it.current()->x() * del)) ; 
    int qy = (int) ((double)(it.current()->y() * del)) ;
    it.current()->move ( qx, qy) ;
  }
  vxsize  *= del ;
  vysize  *= del ;
  emit resized(vxsize, vysize) ;
  repaint () ;
}

void IconDesk::initDrag (Icon *ic)
{
  ic->draw (x11Display (), rootDesk) ;
}

void IconDesk::resizeEvent (QResizeEvent *re)
{
  (void) re ;
  if (width()>vxsize)
    vxsize = width () ;
  if (height() > vysize)
    vysize = height () ;
  
  emit resized (vxsize, vysize) ;
}

void IconDesk::horizScroll (int h)
{
  offsetx = -h * 50 ;
  // scroll (-h, 0) ;
  update () ;
}

void IconDesk::verticScroll (int v)
{
  offsety = -v * 50;
  // scroll (-h, 0) ;
  update () ;
}

void IconDesk::unselectAll ()
{
  /* unselect all icons 
   * and update display
   */

  if (selectedList->isEmpty())
    return ;
  QListIterator<Icon> it (*selectedList) ;
  
  for ( ; it.current() ; ++it) {
    it.current()->set_selected (false) ;
    single_repaint (it.current()) ;
    //    repaint (it.current()->rect(), false) ;
  }
  selectedList->clear() ;
}

void IconDesk::select (Icon* ic)
{
  /* add this icon to selected list
   * and update display
   */
  selectedList->append (ic) ;
  ic->set_selected (true) ;
  single_repaint (ic) ;
}


int IconDesk::appropSize () const 
{
  /* choose a font size based upon the 
   * current scaling 
   */

  return scale+5 ;
}

/* 
 * --------------------------------------------------------------
 * convenience routines
 * -------------------------------------------------------------
 */

Icon *IconDesk::point_is_in_icon (const QPoint &pos) const
{
  /* given a point p in the widget local coordinates, return
   * the icon that the point is in, or 0 if it is not in an icon
   */
  
  // offset to origin
  
  QPoint p (pos.x()-offsetx, pos.y()-offsety) ;

  QListIterator<Icon> it (*icon_list) ;
  
  for ( ; it.current() ; ++it) {
    
    if (p.x() > it.current()->x() && 
	p.y() > it.current()->y() &&
	p.x() < it.current()->x()+it.current()->width() &&
	p.y() < it.current()->y()+it.current()->height() ) {
      // we are in the icon rectangle
      
      return it.current() ;
    }
  }
  
  return 0 ;
}

bool IconDesk::point_is_in_icon (const QPoint &pos, const Icon* ic) const
{
  /* given a point p in the widget local coordinates, return
   * true if the point is in the Icon ic
   */
  
  // allow for scrolling offset
  
  QPoint p (pos.x()-offsetx, pos.y()-offsety) ;
  
  if (p.x() > ic->x() && 
      p.y() > ic->y() &&
      p.x() < ic->x()+ic->width() &&
      p.y() < ic->y()+ic->height() ) {
    // we are in the icon rectangle
    
    return true ;
    }
  
  
  return false ;
}

Icon *IconDesk::point_is_in_icon_label (const QPoint &pos) const
{
  /* given a point p in the widget local coordinates, return
   * the icon whose label the point is in, or 0 if it is not in an icon label
   */

  QListIterator<Icon> it (*icon_list) ;

  // allow for scrolling offset
  QPoint p (pos.x()-offsetx, pos.y()-offsety) ;  

  for ( ; it.current() ; ++it) {  
    if (it.current()->get_label_rect().contains (p))
      return it.current() ;
  }
  return 0 ;
}

void IconDesk::rename_by_label ()
{
  /*
   * callback for the label edit widget
   */

  // if this is called, only
  // one widget should be selected

  if (selectedList->count() != 1) {
    debug ("Bad selected list count (not one) at rename") ;
    exit (1) ;
  }
  
  const QFileInfo *oldf = selectedList->getFirst()->get_file_info () ;

  if (rename (selectedList->getFirst()->get_label(), labelEdit->text()) ) {
    perror ("rename") ;
  } else {
    QFileInfo newf = QFileInfo (oldf->dir(), labelEdit->text()) ;
    emit broadcast_file_change (*oldf, newf) ;
    emit signal_refresh () ;
  }
  delete labelEdit ; labelEdit = 0 ;
}
  
void IconDesk::slot_refresh ()
{
  /*
   * check to make sure contents of desk are
   * consistent with disk
   */
  
  // exploratory code - really inefficient

  /* Algorithm:
   * go through a list 
   */
  
  if (current_dir == 0) return ;

  struct stat buffer ;
  struct stat fbuffer ;
  DIR *dir_pointer ;
  struct dirent *dp ;

  // check mod date on directory
  if (lstat (current_dir->absPath(), &buffer)) {
    perror ("lstat") ;
    return ;
  }

  if (dir_last_mod >= buffer.st_mtime) return ;
  //  debug ("changes...") ;
  dir_last_mod = buffer.st_mtime ;
  
  QList<QString> new_names ;
  QListIterator<QString> new_names_it (new_names) ;

  // save current dir
  QDir cur = QDir::current() ;
  QDir::setCurrent(current_dir->absPath()) ;

  // ok, now has a file been added?
  
  dir_pointer = opendir (".") ;
  while (dp = readdir (dir_pointer)) {
    if (!strcmp (".", dp->d_name) || 
	!strcmp ("..", dp->d_name)) 
      continue ;
    // it would be quicker just to check modification times
    // however, some programs (in particular, gzip)
    // preserve these times: we would miss the creation of a new file
    /*
    if (dir_last_mod <= fbuffer.st_atime) {
    if (lstat (dp->d_name, &fbuffer)) {
      perror ("lstat") ;
      return ;
      }

      debug ("A possible new file %s",dp->d_name) ;
      */
    // dot files ?
    if (*(dp->d_name) == '.') 
      continue ;
    new_names.append (new QString (dp->d_name)) ;
    //    }
  }
  closedir (dir_pointer) ;


  // check to see if a file has been removed

  QListIterator<Icon> it (*icon_list) ;
  for ( ; it.current() ; ) {
    
    // check that it is not in the list of new names
    new_names_it.toFirst () ;
    for ( ; new_names_it.current() ; ++new_names_it) {
      if (*new_names_it.current() == it.current()->get_label())
	new_names.removeRef (new_names_it.current()) ;
    }
    
    if ( ! QFile::exists (it.current()->get_label()) ) {
      
      //      debug ("Icon %s is stale", it.current()->get_label()) ;
      update (it.current()->x(), it.current()->y(),
	      it.current()->width(),it.current()->height()) ;
      const QRect r = it.current()->get_label_rect() ;
      update (r.x(), r.y(), r.width(), r.height ()) ;
      remove_icon (it.current()) ;
    } else {
      ++it ;
    }
  }
  
  // add new icons for remaining new names

  QRegExp filt (current_dir->nameFilter(), true, true) ;
  //  debug ("filt pattern %s", filt.nameFilter()) ;
  new_names_it.toFirst () ;
  for ( ; new_names_it.current() ; ++new_names_it) {    
    // if user has a dirbar regexp set, check that 
    // it matches
    if (filt.match (*new_names_it.current()) >= 0) {
      add_new_icon (*new_names_it.current()) ;
      delete new_names_it.current() ;
    }
  }

  QDir::setCurrent(cur.absPath()) ;

  return ;

}  

void IconDesk::slot_file_change (const QFileInfo &f, const QFileInfo &newf)
{
  /*
   * a signal has been broadcast that the file with
   * (abs) name f has changed name or directory
   */
  
  (void) &f ; (void) &newf ;
  
  if (!current_dir) 
    return ; // an Application window, names can't change
  
  // is it in our directory 
  //  debug ("my dir is %s", (const char *)current_dir->absPath()) ;
  // debug ("changed file is in %s", (const char  *)(f.dirPath(true))) ;
  
  if (!strcmp(current_dir->absPath(),f.dirPath(true))) { // canonical forms?
    
    // it is here, find out what Icon has
    // this label
    
    QListIterator<Icon> it (*icon_list) ;
    for ( ; it.current() ; ++it) {
      if (it.current()->get_label() == f.fileName()) {
	it.current()->new_file_info (newf) ;
	single_repaint (it.current()) ;
	return ;
      }
    }
    debug ("Inconsistancy looking for %s", (const char *)f.fileName()) ;
  }
}
  
		       
void IconDesk::add_new_icon (const DeskIcon &ic) 
{
  // put a new icon into the list, using the data and position
  // of the desk icon ic (in global coordinates)

  icon_list->append ( new Icon(ic.get_icon())) ;
  icon_list->current()->set_accepts_drag_drop (icon_list->current()->is_drag_droppable()) ;
  
  icon_list->current()->move (mapFromGlobal (ic.pos())) ;
}

void IconDesk::add_new_icon (const char *c)
{
  // put a new icon into the list, name c, 
  // at an appropriate position
  
  Icon *nic = new Icon (c, 0, scale) ;
  //  nic->move (20,20) ;
  move_to_empty_space (*nic) ;
  icon_list->append (nic) ;  
  update (nic->x(), nic->y(),
	  nic->width(),nic->height()) ;
  
  
  
}

void IconDesk::remove_icon (const Icon *ic)
{
  // remove this icon from everywhere

  icon_list->remove (ic) ;
  selectedList->remove (ic) ;

  if (focusIcon == ic)
    focusIcon = 0 ;
  if (dragDropIcon == ic)
    dragDropIcon = 0 ;
}

void IconDesk::save_locations () 
{
  /* save locations of icons, if this is our directory
   */

  

  icon_it->toFirst () ;
  for ( ; icon_it->current() ; ++*icon_it) {
    
  }
}  
  
void IconDesk::move_to_empty_space (Icon &ic)
{
  // finds a space to put a new icon

  for (int j = INSET ; j < vysize ; 
       j += INSET + INSET/2 + fontMetrics().lineSpacing()) {
    for (int i = INSET ; i < vxsize ; i += INSET) {
      if (rect_meets_icon (QRect (i,j,ic.width(),ic.height())))
	continue ;

      ic.move (i,j) ;
      return ;
    }
  }
  ic.move (70,70) ;
}

bool IconDesk::rect_meets_icon (const QRect &r)
{

  // check if the rect intersects any icons
  
  QListIterator<Icon> it (*icon_list) ;
  
  for ( ; it.current() ; ++it) {
    
    if (r.intersects (it.current()->rect())) 
      return true ;
  }
  return false ;
}
